// Generator for Superhero and Wobblerne 2.
// https://www.youtube.com/watch?v=3ORZe2FwHyQ
// https://www.youtube.com/watch?v=rtrriyFqFDw
// Written by Nils Liaaen Corneliusen 2019.
// License: CC0 1.0 Universal (CC0 1.0) Public Domain Dedication license
// https://www.ignorantus.com


// Superhero
// ffmpeg -framerate 50 -i test_%05d.jpg -framerate 50 -c:v libx264 -preset slow -crf 18 -pix_fmt yuv420p -s 1920x1080 temp_test.mp4
// ffmpeg -framerate 50 -i test_%05d.jpg -i C:\Users\ncorn\soup\Superhero2_4.wav -framerate 50 -c:v libx264 -preset slow -crf 18 -pix_fmt yuv420p -s 1920x1080 -movflags +faststart Superhero.mp4

// W2
// ffmpeg -framerate 50 -i w2_%05d.jpg -framerate 50 -c:v libx264 -preset fast -crf 18 -pix_fmt yuv420p -s 1920x1080 w2.mp4
// ffmpeg -framerate 50 -i w2_%05d.jpg -i C:\Users\ncorn\w2\wobblerne2.wav -framerate 50 -c:v libx264 -preset slow -crf 18 -pix_fmt yuv420p -s 1920x1080 -movflags +faststart Wobblerne2.mp4


#include "bmp.h"
#include "coeffs.h"
#include "scale.h"

#include <iostream>
#include <cmath>
#include <vector>
#include <stdint.h>
#include <jpeglib.h>
#include <Windows.h>

using namespace std;

#define RAW_FILE "C:\\users\\ncorn\\soup\\raw.yuv"

#define SUPER_BMP "C:\\users\\ncorn\\soup\\superhero.bmp"

#define W2_BMP "C:\\users\\ncorn\\w2\\wobblerne2.bmp"

#define OUT_DIR_SUP "C:\\users\\ncorn\\soup_frames"

#define OUT_DIR_WOB "C:\\users\\ncorn\\w2_frames"

#define WIDTH 720
#define HEIGHT 576
#define PLANE_SIZE (WIDTH*HEIGHT)
#define FRAME_SIZE (PLANE_SIZE*3)
#define FRAMES 30910

#define FRAME_NUMBER(n) (FRAME_SIZE*(n))

#define MIN(a,b) ((a)<(b)?(a):(b))


void resample( uint8_t *src, int srcw, int srch, int srcstr )
{
	for( int y = 0; y < srch; y++ ) {
	
		uint8_t *row = src + srcw*y;

		for( int x = 0; x < srcw; x++ ) {
			double b = *row;
			
			if( b < 16.0 ) b = 16.0;
			if( b > 240.0 ) b = 240.0;

			b -= 16.0;
			b *= 1.138392857142857;
			b = round( b );
			if( b > 255.0 ) {
				printf( "eeeek %f\n", b );
				exit( 1 );
			}
			uint8_t ub = (uint8_t)b;

			*row++ = ub;
		}
	
	}

}


bool saveJPEG(uint8_t *src, char *fname, int quality)
{
    FILE *fp = fopen( fname, "wb" );
    if( fp == NULL ) {
        printf( "Could not create ouput file %s\n", fname );
        return false;
    }

	// More info on YUV420P encoding -> 3.11 Raw (downsampled) image data
	// https://www4.cs.fau.de/Services/Doc/graphics/doc/jpeg/libjpeg.html
	// https://stackoverflow.com/questions/16390783/how-to-compress-yuyv-raw-data-to-jpeg-using-libjpeg

	jpeg_compress_struct cinfo = {0};
	jpeg_error_mgr jerr = {0};

	cinfo.err = jpeg_std_error(&jerr);

    jpeg_create_compress( &cinfo );

    jpeg_stdio_dest(&cinfo, fp);

	cinfo.image_width = 1920; /* Four fields of the cinfo struct must be filled in: */
	cinfo.image_height = 1080;

	cinfo.in_color_space = JCS_YCbCr;
	cinfo.input_components = 3;

	jpeg_set_defaults(&cinfo);
	jpeg_set_quality(&cinfo, quality, TRUE);

	// yuv444
    cinfo.comp_info[0].v_samp_factor = 1;
    cinfo.comp_info[0].h_samp_factor = 1;

	cinfo.raw_data_in = true;
	jpeg_start_compress(&cinfo, true);

	uint8_t *yplane = src;
	uint8_t *uplane = src + 1920*1080;
	uint8_t *vplane = src + 1920*1080*2;

	JSAMPROW y_row[16];
	JSAMPROW u_row[16];
	JSAMPROW v_row[16];

	JSAMPARRAY rows[]  { y_row, u_row, v_row };

	while(cinfo.next_scanline < cinfo.image_height) {

		for( int i = 0; i < 16; i++ ) {
			y_row[i] = yplane + 1920*MIN( cinfo.next_scanline+i, cinfo.image_height-1 );
			u_row[i] = uplane + 1920*MIN( cinfo.next_scanline+i, cinfo.image_height-1 );
			v_row[i] = vplane + 1920*MIN( cinfo.next_scanline+i, cinfo.image_height-1 );
		}

		int rowproc = jpeg_write_raw_data(&cinfo, rows, 16);
	}

	jpeg_finish_compress(&cinfo);
	jpeg_destroy_compress(&cinfo);

	fclose( fp );

	return true;
}


uint8_t *loadfile( const char *fname, uint64_t startframe, uint64_t endframe )
{
	uint64_t offset = startframe*FRAME_SIZE;
	uint64_t length = (endframe-startframe+1)*FRAME_SIZE;

	printf( "start: %zu end: %zu offset: %zu length %zu\n", startframe, endframe, offset, length );

	FILE *fp = fopen( fname, "rb" );
	if( fp == NULL ) {
		printf( "could not open file %s\n", fname );
		return NULL;
	}

	int rv = _fseeki64( fp, offset, SEEK_SET );
	if( rv != 0 ) {
		printf( "doh\n" );
		return NULL;
	}

	uint8_t *buf = (uint8_t *)malloc( length );

	size_t rc = fread( buf, 1, length, fp );
	fclose( fp );
	if( rc != length ) {
		printf( "%s: short read. got %zu, expected %zu\n", fname, rc, length );
		free( buf );
		return NULL;
	}

	printf( "loaded %s: %zu bytes\n", fname, length );
	return buf;
}

void blitit( uint8_t *src, int srcx, int srcy, int srcw, int srch, uint8_t *dst )
{
	for( int y = 0; y < srch; y++ ) {
	
		for( int x = 0; x < srcw; x++ ) {

			*(dst+srcw*(y)+(x)) = *(src+WIDTH*(y+srcy)+x+srcx);
		
		}
	
	}

}

static float scalefac = 0.65f;


// 04:20-09:14
// secs: 260-554
// frames: 13000-27700
// 14700

void genW2( void )
{
	int phys_start =  13000;
	int phys_end   =  27700;

	uint64_t start = 0;
	uint64_t end = phys_end-phys_start;

	uint8_t *data = loadfile( RAW_FILE, phys_start, phys_end );
	if( data == NULL ) {
		printf( "doh\n" );
		return;
	}

	int cropb = 10;
	int cropt = 6; // 5
	int cropl = 6;
	int cropr = 4; // 3

/*
	int cropt = 16;
	int cropb = 16;
	int cropl = 16;
	int cropr = 16;
*/
	int cropw = WIDTH-cropr-cropl;
	int croph = HEIGHT-cropb-cropt;

	uint8_t *cropframe = (uint8_t *)malloc( cropw*croph*3 );

	uint8_t *scaleframe = (uint8_t *)malloc( 1920*1080*3 );

	printf( "%d,%d\n", cropw, croph );

	uint32_t cox[COEFFS_PHASES];
	uint32_t coy[COEFFS_PHASES];

	coeffs_get( scalefac, cox );
	coeffs_get( scalefac, coy );

	uint64_t framectr = 0;

	bmp_argb *bp = bmp_load( W2_BMP, true );

	for( int i = 0; i < 200; i++ ) {
		char fname[128];
		sprintf( fname, "%s\\w2_%05zu.jpg", OUT_DIR_WOB, framectr );
		bmp_save_jpg( bp, fname, false, 100 );
		framectr++;
	}

	for(uint64_t i = start; i <= end; i++ ) {

		blitit( data+FRAME_NUMBER(i)+PLANE_SIZE*0, cropl, cropt, cropw, croph, cropframe+cropw*croph*0 );
		blitit( data+FRAME_NUMBER(i)+PLANE_SIZE*1, cropl, cropt, cropw, croph, cropframe+cropw*croph*1 );
		blitit( data+FRAME_NUMBER(i)+PLANE_SIZE*2, cropl, cropt, cropw, croph, cropframe+cropw*croph*2 );

		resample( cropframe, cropw, croph, cropw );

		char fname[128];
		sprintf( fname, "%s\\w2_%05zu.jpg", OUT_DIR_WOB, framectr );

		printf( "%s\n", fname );

		//1350x1080

		scale_plane( cropframe+cropw*croph*0, cropw, croph, cropw, scaleframe+1920*1080*0+(1920-1350)/2, 1350, 1080, 1920, cox, coy );
		scale_plane( cropframe+cropw*croph*1, cropw, croph, cropw, scaleframe+1920*1080*1+(1920-1350)/2, 1350, 1080, 1920, cox, coy );
		scale_plane( cropframe+cropw*croph*2, cropw, croph, cropw, scaleframe+1920*1080*2+(1920-1350)/2, 1350, 1080, 1920, cox, coy );

		for( int y = 0; y < 1080; y++ ) {
		
			for( int x = 0; x < (1920-1350)/2; x++ ) {

				*(scaleframe+y*1920+x+1920*1080*0) = 0x00;
				*(scaleframe+y*1920+x+1920*1080*1) = 0x80;
				*(scaleframe+y*1920+x+1920*1080*2) = 0x80;

			}

			for( int x = 1350+(1920-1350)/2; x < 1920; x++ ) {

				*(scaleframe+y*1920+x+1920*1080*0) = 0x00;
				*(scaleframe+y*1920+x+1920*1080*1) = 0x80;
				*(scaleframe+y*1920+x+1920*1080*2) = 0x80;

			}
		
		}

/*
		scale_plane( data+FRAME_NUMBER(i)+PLANE_SIZE*0, cropw, croph, cropw, scaleframe+1920*1080*0, 1920, 1080, 1920, cox, coy );
		scale_plane( data+FRAME_NUMBER(i)+PLANE_SIZE*1, cropw, croph, cropw, scaleframe+1920*1080*1, 1920, 1080, 1920, cox, coy );
		scale_plane( data+FRAME_NUMBER(i)+PLANE_SIZE*2, cropw, croph, cropw, scaleframe+1920*1080*2, 1920, 1080, 1920, cox, coy );
*/
//		memset( scaleframe+1920*1080*1, 0x80, 1920*1080 );
//		memset( scaleframe+1920*1080*2, 0x80, 1920*1080 );

		saveJPEG( scaleframe, fname, 100 );

		framectr++;

	}

}




// 50 fps
// 00:03-03:50 Superhero sec 3-230
// rev 1 frame 345-11500
// rev 2 frame 355-11350 10995 frames

// remember 50fps!!!!

void genSouperhero( void )
{
	int phys_start =    355;
	int phys_end   =  11500;

	uint64_t start = 0;
	uint64_t end = phys_end-phys_start;

	uint8_t *data = loadfile( RAW_FILE, phys_start, phys_end );

	int cropb = 10;
	int cropt = 6; // 5
	int cropl = 6;
	int cropr = 4; // 3

/*
	int cropt = 16;
	int cropb = 16;
	int cropl = 16;
	int cropr = 16;
*/
	int cropw = WIDTH-cropr-cropl;
	int croph = HEIGHT-cropb-cropt;

	uint8_t *cropframe = (uint8_t *)malloc( cropw*croph*3 );

	uint8_t *scaleframe = (uint8_t *)malloc( 1920*1080*3 );

	printf( "%d,%d\n", cropw, croph );

	uint32_t cox[COEFFS_PHASES];
	uint32_t coy[COEFFS_PHASES];

	coeffs_get( scalefac, cox );
	coeffs_get( scalefac, coy );

	uint64_t framectr = 0;

	bmp_argb *bp = bmp_load( SUPER_BMP, true );

	for( int i = 0; i < 200; i++ ) {
		char fname[128];
		sprintf( fname, "%s\\test_%05zu.jpg", OUT_DIR_SUP, framectr );
		bmp_save_jpg( bp, fname, false, 100 );
		framectr++;
	}

	for(uint64_t i = start; i <= end; i++ ) {

		blitit( data+FRAME_NUMBER(i)+PLANE_SIZE*0, cropl, cropt, cropw, croph, cropframe+cropw*croph*0 );
		blitit( data+FRAME_NUMBER(i)+PLANE_SIZE*1, cropl, cropt, cropw, croph, cropframe+cropw*croph*1 );
		blitit( data+FRAME_NUMBER(i)+PLANE_SIZE*2, cropl, cropt, cropw, croph, cropframe+cropw*croph*2 );

		resample( cropframe, cropw, croph, cropw );

		char fname[128];
		sprintf( fname, "%s\\test_%05zu.jpg", OUT_DIR_SUP, framectr );

		printf( "%s\n", fname );

		//1350x1080

		scale_plane( cropframe+cropw*croph*0, cropw, croph, cropw, scaleframe+1920*1080*0+(1920-1350)/2, 1350, 1080, 1920, cox, coy );
		scale_plane( cropframe+cropw*croph*1, cropw, croph, cropw, scaleframe+1920*1080*1+(1920-1350)/2, 1350, 1080, 1920, cox, coy );
		scale_plane( cropframe+cropw*croph*2, cropw, croph, cropw, scaleframe+1920*1080*2+(1920-1350)/2, 1350, 1080, 1920, cox, coy );

		for( int y = 0; y < 1080; y++ ) {
		
			for( int x = 0; x < (1920-1350)/2; x++ ) {

				*(scaleframe+y*1920+x+1920*1080*0) = 0x00;
				*(scaleframe+y*1920+x+1920*1080*1) = 0x80;
				*(scaleframe+y*1920+x+1920*1080*2) = 0x80;

			}

			for( int x = 1350+(1920-1350)/2; x < 1920; x++ ) {

				*(scaleframe+y*1920+x+1920*1080*0) = 0x00;
				*(scaleframe+y*1920+x+1920*1080*1) = 0x80;
				*(scaleframe+y*1920+x+1920*1080*2) = 0x80;

			}
		
		}

/*
		scale_plane( data+FRAME_NUMBER(i)+PLANE_SIZE*0, cropw, croph, cropw, scaleframe+1920*1080*0, 1920, 1080, 1920, cox, coy );
		scale_plane( data+FRAME_NUMBER(i)+PLANE_SIZE*1, cropw, croph, cropw, scaleframe+1920*1080*1, 1920, 1080, 1920, cox, coy );
		scale_plane( data+FRAME_NUMBER(i)+PLANE_SIZE*2, cropw, croph, cropw, scaleframe+1920*1080*2, 1920, 1080, 1920, cox, coy );
*/
//		memset( scaleframe+1920*1080*1, 0x80, 1920*1080 );
//		memset( scaleframe+1920*1080*2, 0x80, 1920*1080 );

		saveJPEG( scaleframe, fname, 100 );

		framectr++;

	}

}


int main()
{
	genSouperhero();

//	genW2();

    return 0;
}
